/* Copyright 2015 The TensorFlow Authors. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================*/

// See docs in ../ops/math_ops.cc.

#define EIGEN_USE_THREADS

#include "third_party/eigen3/Eigen/Core"
#include "third_party/eigen3/unsupported/Eigen/CXX11/Tensor"
#include <vector>
#include "tensorflow/core/framework/bounds_check.h"
#include "tensorflow/core/framework/numeric_op.h"
#include "tensorflow/core/framework/op_kernel.h"
#include "tensorflow/core/framework/register_types.h"
#include "tensorflow/core/framework/tensor.h"
#include "tensorflow/core/framework/tensor_types.h"
#include "tensorflow/core/framework/types.h"
#include "tensorflow/core/kernels/segment_reduction_ali_ops_util.h"
#include "tensorflow/core/lib/core/status.h"
#include "tensorflow/core/platform/logging.h"
#include "tensorflow/core/util/util.h"
#include "tensorflow/core/util/work_sharder.h"

namespace tensorflow {

typedef Eigen::ThreadPoolDevice CPUDevice;

// ____________________________________________________________________________
// Sparse segment reduction ops.

// Same as SegmentReductionOp but takes as input a "sparse" tensor, represented
// by two dense tensors, one containing the data, and the other containing
// indices into the data.
template <typename Device, class T>
class SparseSegmentReductionAliOpBase : public OpKernel {
 public:
  explicit SparseSegmentReductionAliOpBase(OpKernelConstruction* context,
                                          bool is_mean, bool is_sqrtn,
                                          bool has_num_segments,
                                          T default_value)
      : OpKernel(context),
        has_num_segments_(has_num_segments),
        reducer_(is_mean, is_sqrtn, has_num_segments, default_value) {}

  void Compute(OpKernelContext* context) override {
    const Tensor& input = context->input(0);
    const Tensor& indices = context->input(1);
    const Tensor& segment_ids = context->input(2);
    
    Tensor num_segments;
    if (has_num_segments_) {
      num_segments = context->input(3);
    }

    Tensor output;
    reducer_.Reduce(context, input, indices, segment_ids, num_segments,
        context->output_alloc_attr(0), &output);
    context->set_output(0, output);
  }

 private:
  const bool has_num_segments_;
  SparseSegmentReduction<Device, T> reducer_;
};

template <typename Device, class T>
class SparseSegmentReductionMeanAliOp
    : public SparseSegmentReductionAliOpBase<Device, T> {
 public:
  explicit SparseSegmentReductionMeanAliOp(OpKernelConstruction* context)
      : SparseSegmentReductionAliOpBase<Device, T>(
            context, true /*is_mean*/, false /*is_sqrtn*/,
            false /* has_num_segments */, T(0) /* default_value */) {}
};

template <typename Device, class T>
class SparseSegmentReductionMeanWithNumSegmentsAliOp
    : public SparseSegmentReductionAliOpBase<Device, T> {
 public:
  explicit SparseSegmentReductionMeanWithNumSegmentsAliOp(
      OpKernelConstruction* context)
      : SparseSegmentReductionAliOpBase<Device, T>(
            context, true /*is_mean*/, false /*is_sqrtn*/,
            true /* has_num_segments */, T(0) /* default_value */) {}
};

template <typename Device, class T>
class SparseSegmentReductionSqrtNAliOp
    : public SparseSegmentReductionAliOpBase<Device, T> {
 public:
  explicit SparseSegmentReductionSqrtNAliOp(OpKernelConstruction* context)
      : SparseSegmentReductionAliOpBase<Device, T>(
            context, false /*is_mean*/, true /*is_sqrtn*/,
            false /* has_num_segments */, T(0) /* default_value */) {}
};

template <typename Device, class T>
class SparseSegmentReductionSqrtNWithNumSegmentsAliOp
    : public SparseSegmentReductionAliOpBase<Device, T> {
 public:
  explicit SparseSegmentReductionSqrtNWithNumSegmentsAliOp(
      OpKernelConstruction* context)
      : SparseSegmentReductionAliOpBase<Device, T>(
            context, false /*is_mean*/, true /*is_sqrtn*/,
            true /* has_num_segments */, T(0) /* default_value */) {}
};

template <typename Device, class T>
class SparseSegmentReductionSumAliOp
    : public SparseSegmentReductionAliOpBase<Device, T> {
 public:
  explicit SparseSegmentReductionSumAliOp(OpKernelConstruction* context)
      : SparseSegmentReductionAliOpBase<Device, T>(
            context, false /*is_mean*/, false /*is_sqrtn*/,
            false /* has_num_segments */, T(0) /* default_value */) {}
};

template <typename Device, class T>
class SparseSegmentReductionSumWithNumSegmentsAliOp
    : public SparseSegmentReductionAliOpBase<Device, T> {
 public:
  explicit SparseSegmentReductionSumWithNumSegmentsAliOp(
      OpKernelConstruction* context)
      : SparseSegmentReductionAliOpBase<Device, T>(
            context, false /*is_mean*/, false /*is_sqrtn*/,
            true /* has_num_segments */, T(0) /* default_value */) {}
};

#define REGISTER_CPU_SPARSE_KERNELS(type)                                   \
  REGISTER_KERNEL_BUILDER(Name("SparseSegmentSum")                          \
                              .Device(DEVICE_CPU)                           \
                              .TypeConstraint<type>("T")                    \
                              .TypeConstraint<int32>("Tidx"),               \
                          SparseSegmentReductionSumAliOp<CPUDevice, type>); \
  REGISTER_KERNEL_BUILDER(                                                  \
      Name("SparseSegmentSumWithNumSegments")                               \
          .Device(DEVICE_CPU)                                               \
          .TypeConstraint<type>("T")                                        \
          .TypeConstraint<int32>("Tidx"),                                   \
      SparseSegmentReductionSumWithNumSegmentsAliOp<CPUDevice, type>);
TF_CALL_REAL_NUMBER_TYPES(REGISTER_CPU_SPARSE_KERNELS);
#undef REGISTER_CPU_SPARSE_KERNELS

#define REGISTER_CPU_SPARSE_KERNELS(type)                                    \
  REGISTER_KERNEL_BUILDER(Name("SparseSegmentMean")                          \
                              .Device(DEVICE_CPU)                            \
                              .TypeConstraint<type>("T")                     \
                              .TypeConstraint<int32>("Tidx"),                \
                          SparseSegmentReductionMeanAliOp<CPUDevice, type>); \
  REGISTER_KERNEL_BUILDER(                                                   \
      Name("SparseSegmentMeanWithNumSegments")                               \
          .Device(DEVICE_CPU)                                                \
          .TypeConstraint<type>("T")                                         \
          .TypeConstraint<int32>("Tidx"),                                    \
      SparseSegmentReductionMeanWithNumSegmentsAliOp<CPUDevice, type>);
REGISTER_CPU_SPARSE_KERNELS(float);
REGISTER_CPU_SPARSE_KERNELS(double);
#undef REGISTER_CPU_SPARSE_KERNELS

#define REGISTER_CPU_SPARSE_KERNELS(type)                                     \
  REGISTER_KERNEL_BUILDER(Name("SparseSegmentSqrtN")                          \
                              .Device(DEVICE_CPU)                             \
                              .TypeConstraint<type>("T")                      \
                              .TypeConstraint<int32>("Tidx"),                 \
                          SparseSegmentReductionSqrtNAliOp<CPUDevice, type>); \
  REGISTER_KERNEL_BUILDER(                                                    \
      Name("SparseSegmentSqrtNWithNumSegments")                               \
          .Device(DEVICE_CPU)                                                 \
          .TypeConstraint<type>("T")                                          \
          .TypeConstraint<int32>("Tidx"),                                     \
      SparseSegmentReductionSqrtNWithNumSegmentsAliOp<CPUDevice, type>);
REGISTER_CPU_SPARSE_KERNELS(float);
REGISTER_CPU_SPARSE_KERNELS(double);
#undef REGISTER_CPU_SPARSE_KERNELS

template <class T>
class SparseSegmentGradAliOpBase : public OpKernel {
 public:
  explicit SparseSegmentGradAliOpBase(OpKernelConstruction* context,
                                     bool is_sqrtn)
      : OpKernel(context), reducer_(is_sqrtn) {}

  void Compute(OpKernelContext* context) override {
    const Tensor& input = context->input(0);
    const Tensor& indices = context->input(1);
    const Tensor& segment_ids = context->input(2);
    const Tensor& output_dim0 = context->input(3);

    OP_REQUIRES(context, TensorShapeUtils::IsVector(indices.shape()),
                errors::InvalidArgument("indices should be a vector."));
    OP_REQUIRES(context, TensorShapeUtils::IsVector(segment_ids.shape()),
                errors::InvalidArgument("segment_ids should be a vector."));
    OP_REQUIRES(context, IsLegacyScalar(output_dim0.shape()),
                errors::InvalidArgument("output_dim0 should be a scalar."));

    Tensor output;
    reducer_.ReduceGrad<T>(context, input, indices, segment_ids,
        output_dim0, output);
    context->set_output(0, output);
  }

 private:
  SparseSegmentReductionGrad reducer_;
};

template <class T>
class SparseSegmentMeanGradAliOp : public SparseSegmentGradAliOpBase<T> {
 public:
  explicit SparseSegmentMeanGradAliOp(OpKernelConstruction* context)
      : SparseSegmentGradAliOpBase<T>(context, false /*is_sqrtn*/) {}
};

template <class T>
class SparseSegmentSqrtNGradAliOp : public SparseSegmentGradAliOpBase<T> {
 public:
  explicit SparseSegmentSqrtNGradAliOp(OpKernelConstruction* context)
      : SparseSegmentGradAliOpBase<T>(context, true /*is_sqrtn*/) {}
};

#define REGISTER_CPU_SPARSE_KERNELS(type)                     \
  REGISTER_KERNEL_BUILDER(Name("SparseSegmentMeanGrad")       \
                              .Device(DEVICE_CPU)             \
                              .TypeConstraint<type>("T")      \
                              .TypeConstraint<int32>("Tidx"), \
                          SparseSegmentMeanGradAliOp<type>);
REGISTER_CPU_SPARSE_KERNELS(float);
REGISTER_CPU_SPARSE_KERNELS(double);
#undef REGISTER_CPU_SPARSE_KERNELS

#define REGISTER_CPU_SPARSE_KERNELS(type)                     \
  REGISTER_KERNEL_BUILDER(Name("SparseSegmentSqrtNGrad")      \
                              .Device(DEVICE_CPU)             \
                              .TypeConstraint<type>("T")      \
                              .TypeConstraint<int32>("Tidx"), \
                          SparseSegmentSqrtNGradAliOp<type>);
REGISTER_CPU_SPARSE_KERNELS(float);
REGISTER_CPU_SPARSE_KERNELS(double);
#undef REGISTER_CPU_SPARSE_KERNELS
}  // namespace tensorflow
